#ifndef _SEQREADER_H
#define _SEQREADER_H

#include <vector>
#include <string>
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <cstring>
#include <cmath>
#include "const_include.h"
#include "bin_seq.h"
#include "UnitTest.h"

using namespace std;

struct SeqFile {
	SeqType type;
	string file_name;

	bool operator ()(SeqFile &other) {
		if( file_name == other.file_name)
			return true;
		return false;
	}
};

class SeqReader {
	public:
	
		/*
		 * The Constructor with a given file name.  Used to set everything up, determine the number
		 * of sequence files, and create the set of file strings to use.
		 * 
		 * @param fn - the string received from the SeqDriver program.
		 */
		SeqReader(const char* fn);
		SeqReader();
		~SeqReader();

		/*!
		 * use will enable you to use this file after the SeqDriver has already
		 * been initialized.  
		 *
		 * @param fn - the file(s) to use as sequences
		 *				(can be either _int or _prb)
		 */
		void use(const char* fn);
		void use(string &str);
		
		/*
		 * GetNextSequence will return the next available PWM for the sequence.
		 * If it has run out of sequences, it will read more in from memory.
		 *
		 * @return the PWM representing the next sequence.
		 */
		//vector<vector<double> > GetNextSequence();
		//vector<vector<double> > PeekNextSequence();
		Read* GetNextSequence();
		Read* PeekNextSequence();

		void PrintSeq(double &seq);		
		string GetConsensus(vector<vector<double> > &read);
		string GetConsensus(Read* read);
		unsigned int GetNumSeqs();
		
		/*!
		 * Used with the SeqDriver in determining how many files to use...
		 */
		unsigned int GetNumSeqFiles();
		vector<string>& GetSeqFileVector();
		
		static bool Test(ostream &os);

	private:
		void Init(const char* fn);
		void clearReads(Read* readArr[]);
		
		bool get_more_int();
		bool get_more_prb();
		double Q2Prb(double Q);
		float Compare(string, string, int);
		void FixReads(vector<vector<double> > &read);
		//void FixReads(Read &pwm);
		
		void find_type();

		char max_char(vector<double> &chr);
		//char max_char(float chr[4]);
	
		//vector<vector<vector<double> > > reads;
		//Read reads[READ_BUFFER];	//an array of pointers to reads
		Read* reads[READ_BUFFER];	//an array of pointers to reads
		unsigned int num_reads;
		unsigned int seq_counter;

		unsigned int offset;	//The offset in the individual sequence file.
		unsigned int seq_num;	//To keep track which sequence file we are on.
		//const char* filename;
		//SeqType type;
		string this_filename;
		SeqType this_type;
		vector<string> seq_files;	//A list of the sequence files to read.
};

#endif
