#ifndef BIN_SEQ_H
#define BIN_SEQ_H

#include "UnitTest.h"
#include "const_include.h"
#include "Reader.h"
#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <sstream>

using namespace std;

class bin_seq {
	public:
		bin_seq();
		
		unsigned int str2int();
	
		unsigned int str2int(string &a);
		string hash2str(const unsigned int h, int length);
		
		/*
		 * get_hash will get the hashing value for either a given string or
		 * a given location on genome.  The return value of get_hash is a pair
		 * of bool and BIT64_t, representing whether the hash was successful, and,
		 * if it was, the value for the hash.
		 * get_hash will return false only if there is an unknown character found in
		 * the middle of the sequence.  For example, actgNgcta will not be a unique 
		 * 9-mer, so it is not added to the hash.  In addition, when trying to map
		 * back to a genome, if the read has an unknown character in the given K-mer,
		 * it should not map back exactly to any portion of the genome.
		 */
		pair<int,unsigned long> get_hash(string str);
		pair<int,unsigned long> get_hash(const char* str);
		pair<int,unsigned long> get_hash(unsigned int &pos, const unsigned char* genome);
		
		inline unsigned int base2int(char c);
		inline char int2base(unsigned int i);
		
		//double get_align_score(vector< vector<double> > &pwm, string &gen);
		//string get_align_score_w_traceback(const vector< vector<double> > &pwm, const string &consense, const string &gen);
		float get_align_score(Read &read, string &gen);
		string get_align_score_w_traceback(const Read &read, const string &consense, const string &gen);
		int get_align_score(string seq1, string seq2);
		
		static bool Test(ostream &os);
	private:	
		float get_val(char a, char b);
		//double get_val(const vector<double> &a, const char b);
		float get_val(const float* a, const char b);

		int max(int a, int b, int c);
		//double max_dbl(double a, double b, double c);
		float max_flt(float a, float b, float c);
		void print_align(string str1, string align, string str2);
		
};

#endif
