#include <iostream>
#include <unistd.h>
#include <stdio.h>
#include <vector>
#include <string>
#include <sstream>
#include <cstring>
#include <ctime>
#include <cmath>
#include <set>
#include <map>
#include <pthread.h>
#include "const_include.h"
#include "Genome.h"
#include "bin_seq.h"
#include "SeqReader.h"
#include <algorithm>
#include <exception>

struct scored_sequence;

typedef set<scored_sequence*,scored_sequence > seq_set;
typedef map<string,scored_sequence* > seq_map;
//typedef BST<scored_sequence*> seq_set;

unsigned int BISULFITE = 0;
unsigned int ATOG = 0; 
unsigned char INT2BASE[16];
unsigned char bs_CONVERSION[256];
unsigned char gen_CONVERSION[256];

int VERBOSE=0;
int MER_SIZE=9;
unsigned int MAX_MATCHES = 1000;
unsigned int BUFFER_SIZE = 1000000;
unsigned int HASH_SIZE = 262144;	//HASH_SIZE is set as 4^9
unsigned int MAX_HASH_SIZE = 0;
float ALIGN_SCORE = 0;

float GAP = 0;
float MATCH = 0.5;
float MISMATCH = -0.5;

int MAX_GAP = 3;
const char* seq_file = 0;
const char* genome_file = 0;
const char* output_file = "gnumap_out";
const char* pos_matrix = NULL;
char* adaptor = NULL;
int PRINT_FULL=0;
float ALIGN_SCORES[256][4];
bool perc = false;
int SEQ_LENGTH=-1;

float MIN_ADAPTOR_DIFF = 0.75;	//used to define the minimum difference from the given adaptor sequence to the consensus
unsigned int MIN_CHOPPED_BASES = 4;	//if there's less than 4 bases that match the adaptor, we're not going to chop them.
//ADD_SIZE is not used...
int ADD_SIZE = 3;

//I'm not sure what BIN_SIZE means...It's not used anymore--only appearance is in Genome.cpp:365
//int BIN_SIZE = 10;
int BIN_SIZE = 1;
 
int NUM_THREADS = 1;

int bad_files =0;

bin_seq seq;

pthread_attr_t attr;
pthread_mutex_t lock;

void* thread_print(void* t_opts);

/*****************************************************
 * These next functions are used in parsing command-
 * line arguments
 *****************************************************/
int ParseCmdLine(const int argc, const char* argv[]);
void GetParseError(ostream &os, const char* argv[]);
int set_arg_ext(const char* param, const char* assign, int &count);
int set_arg(const char* param, const char* assign, int &count);
/*struct CMDLine {
	int errno;
	int errpos;
} this_cmd;*/
CMDLine this_cmd;
/*****************************************************/

void usage(int has_error, char* errmessage) {
	if(has_error) cout << endl << errmessage << endl ;

	cout << "Usage: gnumap [options] <file_to_parse>\n"
		 << "  -g, --genome=STRING          Genome .fa file(s)\n"
		 << "  -o, --output=STRING          Output file\n"
		 << "  -a, --align_score=DOUBLE     Limit for sequence alignment (default: 0)\n"
		 << "  -p, --percent=NONE           Use percentage when determining alignment score\n"
		 << "  -v, --verbose=INT            Verbose (default=0)\n"
		 << "  -c, --num_proc=INT           Number of processors to run on\n"
		 << "  -m, --mer_size=INT           Mer size (default=9)\n"
		 << "  -B, --buffer=INT             Buffer size\n"
		 << "  -T, --max_match=INT          Maximum number of matches for a given\n"
		 << "                               sequence (default: 1000)\n"
		 << "  -h, --max_hash=INT           Maximum values to store in the hash\n"
		 << "                               (default: none)\n"
		 << "  -S, --subst_file=STRING      Position-Weight Matrix file for DNA\n"
		 << "                               Substitutions\n"
		 << "  -G, --gap_penalty=DOUBLE     Gap Penalty for Substitution Matrix\n"
		 << "  -M, --max_gap=INT            Maximum Number of Gaps to use in Alignment\n"
		 << "  -A, --adaptor=STRING         Adaptor sequence to remove from sequences\n"
		 << "  -0, --print_full             Print locations for the entire sequence, not\n"
		 << "                               just for the beginning.\n"
		 << "  -b, --bs_seq                 Flag to turn on the C to T conversion, used in\n"
		 << "                               bisulfite sequence analysis\n"
		 << "  -d, --a_to_g                 Flag that allows for A to G conversion\n"
		 << endl
		 << "Help options:\n"
		 << "  -?, --help                   Show this help message\n";
	exit(1);
}

/*
 * Because this is used to produce the consensus sequence, we need ambiguity characters...
 */
//inline char max_char(vector<double> &chr) {
inline char max_char(float* chr) {
	if((chr[0] == chr[1]) && (chr[0] == chr[2]) && (chr[0] == chr[3]))
		return 'n';
	if(chr[0] >= chr[1]) {
		if(chr[0] >= chr[2]) {
			if(chr[0] >= chr[3])
				return 'a';
			else
				return 't';
			}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else
				return 't';
		}
	}
	else {
		if(chr[1] >= chr[2]) {
			if(chr[1] >= chr[3])
				return 'c';
			else
				return 't';
		}
		else {
			if(chr[2] >= chr[3])
				return 'g';
			else return 't';
		}
	}

	return 'n';
}
 
//inline string GetConsensus(vector<vector<double> > &read) {
inline string GetConsensus(Read &read) {
	string seq = "";

	for(unsigned int i=0; i<read.length; i++) {
		seq += max_char(read.pwm[i]);
	}

	return seq;
}

//I don't think this function is even used...
//vector<vector<double> > subvector(vector<vector<double> > &vec, int start, int length) {
float** subvector(float** &vec, int start, int length) {
	//vector<vector<double> > to_return;
	float** to_return = new float*[length];

	for(int i=start,j=0; i<start+length; i++,j++) {
		to_return[j] = new float[4];
		for(int k=0; k<4; k++) {
			to_return[j][k] = vec[i][k];
		}
		delete[] vec[i];
	}
	
	delete[] vec;	//we want to free the memory for vec because we're creating a new
					//one, right??
	
	return to_return;
}

struct thread_opts {
	//thread_opts(Genome &g, vector<vector<double> >* s, unsigned int n) :
	thread_opts(Genome &g, Read* s[], unsigned int n) :
		gen(g), seqs(s), num_seqs(n) {}
	
	Genome& gen;
	//vector<vector<double> >* seqs;
	Read** seqs;
	unsigned int num_seqs;
};

struct scored_sequence {
	string sequence;
	float align_score;
	float log_align_score;
	//the bool in this pair represents which strand:
	//	true = positive strand (+)
	//	false = complementary strand (-)
	set<pair<unsigned long,bool> > positions;

	scored_sequence() : sequence("<empty>"), align_score(-1),
		log_align_score(exp(align_score)) {
		positions.clear();
	}
	
	scored_sequence(string seq, float as, unsigned long pos, bool us) :
		sequence(seq), align_score(as) {
		log_align_score = exp(align_score);
		positions.insert(pair<unsigned long,bool>(pos,us));
	}
	scored_sequence(const scored_sequence &other) {
		Init(other);
	}
	
	scored_sequence& operator =(const scored_sequence &other) {
		//if( !(this == other) )
		Init(other);
		return *this;
	}
	
	inline void Init(const scored_sequence &other) {
		positions.clear();
		sequence = other.sequence;
		align_score = other.align_score;
		log_align_score = other.log_align_score;
		positions = other.positions;
	}
	
	inline int compare(scored_sequence *a, scored_sequence *b) {
		if(a == b)
			return 0;
		else {
			if(a > b)
				return 1;
			else
				return -1;
		}
	}
	inline bool operator() (scored_sequence &a, scored_sequence &b) {
		return a.sequence.compare(b.sequence);
	}
	
	//The return value is 1 if the set doesn't contain an element whose sort key matches the parameter key. 
	//0 if the set does contain an element with a matching key.
	inline bool operator() (const scored_sequence *a, const scored_sequence *b) const {
		return (a->sequence != b->sequence);
	}
	inline bool operator() (const float &a, const float &b) {
		return a != b;
	}
	inline bool operator == (scored_sequence &other) {
		if( (sequence == other.sequence) )
			return true;
		return false;
	}
	inline bool operator == (scored_sequence *other) {
		if( (sequence == other->sequence) )
			return true;
		return false;
	}
	inline bool operator< (const scored_sequence &other) const {
		//return align_score < other.align_score;
		return (sequence < other.sequence);
	}
	inline bool operator< (const scored_sequence* other) const {
		//return align_score < other->align_score;
		return (sequence < other->sequence);
	}
	inline bool operator> (const scored_sequence &other) const {
		float my_total = log_align_score / positions.size();
		float other_total = other.log_align_score / other.positions.size();
		return my_total > other_total;
	}
	inline bool operator> (const scored_sequence *other) const {
		float my_total = log_align_score / positions.size();
		float other_total = other->log_align_score / other->positions.size();
		return my_total > other_total;
	}
	
	inline bool is_greater(const scored_sequence &other, float denom) {
		float my_total = log_align_score / denom;
		float other_total = other.log_align_score / denom;
		return my_total > other_total;
	}
	
	bool add_spot(unsigned long pos, bool us) {
		//cout << "inserting: " << pos << endl;
		if( positions.insert( pair<unsigned long,bool>(pos,us) ).second ) {
			//cout << "inserted. \tsize: " << positions.size() << endl;
			return true;
		}
		else {
			//cout << "already in set. \tsize: " << positions.size() << endl;
			return false;
		}
	}
	
	string print_full_w_score(float denom, Genome &gen) const {
		ostringstream temp;
		float total_score = log_align_score / denom;
		temp << align_score << "\t" << log_align_score << "\t" << positions.size() << "\t" << total_score << "\t" << sequence << "\t";
			
		set<pair<unsigned long,bool> >::iterator sit;
		
		for(sit = positions.begin(); sit != positions.end(); sit++)
			temp << "\t" << gen.GetPos(*sit);
			
		return temp.str();
	}
	
	string print_and_score(float denom, Genome &gen) const {
		ostringstream temp;
		float total_score = log_align_score / denom;
		temp << align_score << "\t" << log_align_score << "\t" << positions.size() << "\t" << total_score << "\t" << sequence << "\t";
		
		set<pair<unsigned long,bool> >::iterator sit;
		
		for(sit = positions.begin(); sit != positions.end(); sit++) {
			temp << "\t" << gen.GetPos(*sit);
			gen.AddScore( (*sit).first,total_score );
		}
			
		return temp.str();
	}

	// score will add the log-odds score (adjusted by denom) to each
	// position on the Genome gen.  Additionally, this can be passed
	// a flag to specify the number of positions after the start position
	// to add to (for adding to different lengths of sequences).
	//void score(float denom, Genome &gen, unsigned int len, vector<vector<double> > &pwm) const {
	//void score(float denom, Genome &gen, unsigned int len, Read &read) const {
	void score(float denom, Genome &gen, unsigned int len) const {
		float total_score = log_align_score / denom;
		//set<pair<unsigned int,bool> >::iterator sit;
		set<pair<unsigned long,bool> >::iterator sit;
			
		for(sit=positions.begin(); sit!=positions.end(); sit++)
			for(unsigned int i=0; i<len; i++) 
				gen.AddScore( (*sit).first+i,total_score );
		
	}

	//void score_max_bs(float denom, Genome &gen, unsigned int len, vector<vector<double> > &pwm, ostringstream &str) const {
	void score_max_bs(float denom, Genome &gen, unsigned int len, Read &read, ostringstream &str) const {
		if(!positions.size())	//if there were no matches, just return.
			return;

		float total_score = log_align_score / denom;
		//set<pair<unsigned int,bool> >::iterator sit;

		//we only need to compute this alignment string once because, while there may be many locations for each
		//sequence, we're ensuring that the sequences are the exact same.

		bin_seq bs;
		unsigned int seq_size = sequence.size();

		string read_consensus = GetConsensus(read);

		string gen_string = gen.GetString((*positions.begin()).first, seq_size);
		string aligned = bs.get_align_score_w_traceback(read, read_consensus, gen_string);

		str << aligned << "\t";
	

		set<pair<unsigned long,bool> >::iterator sit;
		for(sit=positions.begin(); sit!=positions.end(); sit++) {
			for(unsigned int i=0; i<aligned.size(); i++) {
				if(gen_string[i] == 'c') {
					switch(aligned[i]) {
						gen.AddGenScore((*sit).first+i,total_score,'c');

						case 'a':
							gen.AddSeqScore((*sit).first+i,total_score,'a');
						break;
						case 'c':
							gen.AddSeqScore((*sit).first+i,total_score,'c');
						break;
						case 'g':
							gen.AddSeqScore((*sit).first+i,total_score,'g');
						break;
						case 't':
							gen.AddSeqScore((*sit).first+i,total_score,'t');
						break;
					}
				}
				/*
				switch(gen_string[i]) {
					case 'a':
						gen.AddGenScore((*sit).first+i,total_score,'a');
					break;
					case 'c':
						gen.AddGenScore((*sit).first+i,total_score,'c');
					break;
					case 'g':
						gen.AddGenScore((*sit).first+i,total_score,'g');
					break;
					case 't':
						gen.AddGenScore((*sit).first+i,total_score,'t');
					break;
				}
				*/
			}
		}
	}
	
	//void score_max_atog(float denom, Genome &gen, unsigned int len, vector<vector<double> > &pwm, ostringstream &str) const {
	void score_max_atog(float denom, Genome &gen, unsigned int len, Read &read, ostringstream &str) const {
		if(!positions.size())
			return;
		float total_score = log_align_score / denom;
		//set<pair<unsigned int,bool> >::iterator sit;
	
		bin_seq bs;
		unsigned int seq_size = sequence.size();

		if(!positions.size())	//if there were no matches, just return.
			return;

		string read_consensus = GetConsensus(read);

		string gen_string = gen.GetString((*positions.begin()).first, seq_size);
		string aligned = bs.get_align_score_w_traceback(read,read_consensus,gen_string);

		str << aligned << "\t";

		set<pair<unsigned long,bool> >::iterator sit;
		for(sit=positions.begin(); sit!=positions.end(); sit++) {
			//string gen_string = gen.GetString((*sit).first, seq_size);
			//string aligned = bs.get_align_score_w_traceback(read, sequence, gen_string);
			//if(!aligned.equals(prev_aligned))
			//	printf("Alignments are Not Equal!\n");
			
			for(unsigned int i=0; i<aligned.size(); i++) {
				if(gen_string[i] == 'a') {
					gen.AddGenScore((*sit).first+i,total_score,'a');
					switch(aligned[i]) {
						case 'a':
							gen.AddSeqScore((*sit).first+i,total_score,'a');
						break;
						case 'c':
							gen.AddSeqScore((*sit).first+i,total_score,'c');
						break;
						case 'g':
							gen.AddSeqScore((*sit).first+i,total_score,'g');
						break;
						case 't':
							gen.AddSeqScore((*sit).first+i,total_score,'t');
						break;
					}
				}
/*
				switch(gen_string[i]) {
					case 'a':
						gen.AddGenScore((*sit).first+i,total_score,'a');
					break;
					case 'c':
						gen.AddGenScore((*sit).first+i,total_score,'c');
					break;
					case 'g':
						gen.AddGenScore((*sit).first+i,total_score,'g');
					break;
					case 't':
						gen.AddGenScore((*sit).first+i,total_score,'t');
					break;
				}
*/
			}
		}
	}
};

inline string reverse_comp(string &s) {
	string temp = "";
	int len = s.length();
	
	for(int i=1; i<=len; i++) {
		switch(s[len-i]) {
			case 'a':
				temp += 't';
				break;
			case 't':
				temp += 'a';
				break;
			case 'c':
				temp += 'g';
				break;
			case 'g':
				temp += 'c';
				break;
			default:
				temp += 'n';
				break;
		}
	}
	
	return temp;
} 

//we modify the former one
//inline vector<vector<double> > reverse_comp(vector<vector<double> > &v) {
inline void reverse_comp(float** pwm, unsigned int length) {
	float temp_pwm[length][4];


	//for(r_vit = v.rbegin(); r_vit != v.rend(); ++r_vit) {
	for(unsigned int i = 0; i < length; ++i) {
		//need to switch places with the characters: a-t, g-c
		//also need to rev-com it.
		temp_pwm[length-1-i][0] = pwm[i][3];
		temp_pwm[length-1-i][1] = pwm[i][2];
		temp_pwm[length-1-i][2] = pwm[i][3];
		temp_pwm[length-1-i][3] = pwm[i][1];
	}
		
	for(unsigned int i=0; i<length; i++) {
		for(unsigned int j=0; j<4; j++) {
			pwm[i][j] = temp_pwm[i][j];
		}
	}
	
	//cout << "temp's size: " << temp.size() << endl;
	//cout << "originals size: " << v.size() << endl;
}

//void clear(seq_set &to_clear) {
void clear(seq_map &to_clear) {
	//seq_set::iterator sit;
	seq_map::iterator sit;
	
	for(sit = to_clear.begin(); sit != to_clear.end(); ++sit) { 
		//delete(*sit);
		delete (*sit).second;
	}
}

//string print_top_matches(Genome &gen, vector<vector<double> > &search, string &consensus) {
string print_top_matches(Genome &gen, Read &search, string &consensus) {
		
	bin_seq bs;
	float this_align_score = 0;
	ostringstream to_return;
	to_return << consensus << "\t";
	if(search.length < (unsigned int)MER_SIZE) { 
		to_return << "Too short to align";
		bad_files++;
		return to_return.str();
	}


	float max_align_score = bs.get_align_score(search,consensus);
	to_return << max_align_score << "\t"; 

	//bad idea:--for testing purposes only.
	//return to_return.str();
	//end of bad idea
	
#ifdef DEBUG
	if(VERBOSE > 2) {
		cout << "\nNext Read:" << endl;
		cout << consensus << "\t" << max_align_score << "\n";
	}
#endif

	if(max_align_score < 0) { 
		bad_files++;
		to_return << "Not possible to obtain a positive alignment score";
		return to_return.str();
	}

	if(perc)
		this_align_score = ALIGN_SCORE * max_align_score;
	else
		this_align_score = ALIGN_SCORE;

	to_return << this_align_score << "\t";
	
	seq_map unique;  
	//seq_set unique;
	//BST<scored_sequence*> unique;
	
	float denominator = 0;
	
	for(unsigned int i=0; i< search.length-MER_SIZE; i++) {
		string consensus_piece = consensus.substr(i,MER_SIZE);
		pair<int, unsigned long> p_hash = bs.get_hash(consensus_piece);
		if(!p_hash.first) 
			continue;

#ifdef DEBUG
		if(VERBOSE > 3)
			to_return << "\nhashing (" << i << "): " << p_hash.second << "\n    " << consensus_piece << endl;
#endif

		vector<unsigned long> hashes = gen.GetMatches(p_hash.second);
		
		vector<unsigned long>::iterator vit;

		for(vit = hashes.begin(); vit != hashes.end(); ++vit) {
			//Match the sequence to a sequence that has two characters before and after.
			//string to_match = gen.GetString((*vit)-i-2, search.size()+4);
			
			unsigned long beginning = ((unsigned long)(*vit)<=(unsigned long)i ? 0 : (*vit-(unsigned long)i));
			
			string to_match = gen.GetString(beginning, search.length);

			float align_score = 0;
			char aligned[to_match.length()];
			strcpy(aligned,to_match.c_str());

			//if(BISULFITE)
				//align_score = bs.get_align_score_w_traceback(search,to_match,aligned);
			//else
				align_score = bs.get_align_score(search,to_match);

#ifdef DEBUG
			if(VERBOSE > 1) 
				cout << "matching: " << to_match << endl
					 << "  with  : " << consensus << "\tat pos   " << beginning 
					 							  << "\tand step " << i
												  << "\tand alignment " << align_score << endl;
#endif
					 				
			if(align_score >= this_align_score) {
				//if(align_score >= (this_align_score + 4.6))
					//this_align_score = align_score-2;

				scored_sequence* temp = new scored_sequence(to_match,align_score,beginning,true);
				//cout << endl << endl;
				pair<seq_map::iterator,bool> it_bool = unique.insert(pair<string,scored_sequence*>(to_match,temp));
				//pair<seq_set::iterator,bool> it_bool = unique.insert(temp);

				if(!it_bool.second) {	//If it wasn't inserted into the set--because 
										//	it had already been inserted...
#ifdef DEBUG
					if(VERBOSE > 3) {
						if( (*it_bool.first)->add_spot((*vit)-i) ) {
							to_return << "score here: " << align_score << ", matches to " << to_match << endl;
							to_return << (*it_bool.first)->print() << endl;
						}
					}
#endif
					
					//int c = unique.count(*it_bool.first);
					//cout << "count: " << c << endl;
					//if(c > 1)
						//cout << "count: " << c << " " << (*it_bool.first)->print() << endl;
					delete temp;
					if( (*it_bool.first).second->add_spot(beginning,true) )
						denominator += exp(align_score);
					//(*it_bool.first)->add_spot((*vit)-i);
				}
				else
					denominator += exp(align_score);
				
			}
		}
		
		if(unique.size() > MAX_MATCHES) {
			clear(unique);
			bad_files++;
			to_return << MAX_MATCHES << "+ (Too many matches to compute.)";
			return to_return.str();
		}
	
	}

	if(BISULFITE || ATOG)
		goto AFTER_REV;

	//We need to do the reverse of this as well...
	consensus = reverse_comp(consensus);
	//search.pwm = reverse_comp(search.pwm,search.length);
	//will modify the search.pwm reference
	reverse_comp(search.pwm,search.length);
	
	for(int i=0; i<(int)search.length-MER_SIZE; i++) {
		string consensus_piece = consensus.substr(i,MER_SIZE);
		pair<bool, unsigned long> p_hash = bs.get_hash(consensus_piece);
		
		//If this doesn't hash uniquely (cannot be any ambiguity characters)
		if(!p_hash.first)
			continue;

		vector<unsigned long> hashes = gen.GetMatches(p_hash.second);
		vector<unsigned long>::iterator vit;

		for(vit = hashes.begin(); vit != hashes.end(); ++vit) {
			//Match the sequence to a sequence that has two characters before and after.
			//string to_match = gen.GetString((*vit)-i-2, search.size()+4);
			unsigned long beginning = ((unsigned long)(*vit)<(unsigned long)i ? 0 : (*vit-(unsigned long)i));
			string to_match = gen.GetString(beginning, search.length);

			float align_score = bs.get_align_score(search,to_match);

			if(align_score >= this_align_score) {
				//if(align_score >= (this_align_score + 4.6))
					//this_align_score = align_score-2;

				scored_sequence* temp = new scored_sequence(to_match,align_score,beginning,false);
				pair<seq_map::iterator,bool> it_bool = unique.insert(pair<string,scored_sequence*>(to_match,temp));

				if(!it_bool.second) {	//If it wasn't inserted into the set--because 
										//	it had already been inserted...
					if( (*it_bool.first).second->add_spot(beginning,false) )
						denominator += exp(align_score);
					delete temp;
				}
				else
					denominator += exp(align_score);
			}
		}
		
		if(unique.size() > MAX_MATCHES) {
			clear(unique);
			bad_files++;
			to_return << MAX_MATCHES << "+ (Too many matches to compute.)";
			return to_return.str();		}
	
	}


AFTER_REV:	//for BISULFITE we don't want to do the reverse compliment--doesn't make sense with RNA

	//to_return << endl;
	//to_return << "Denominator: " << denominator << endl;
	//seq_set::iterator sit;
	seq_map::iterator sit;

	scored_sequence max;

	unsigned int len = PRINT_FULL ? consensus.length() : 1;
	
	pthread_mutex_lock(&lock);
	for(sit = unique.begin(); sit != unique.end(); ++sit) {
		if( (*(*sit).second).is_greater(max,denominator))
			max = *(*sit).second;
		//to_return << "\t" << (*sit).second->print_and_score(denominator,gen) << endl;
		//(*sit).second->score(denominator,gen,len,search);
		(*sit).second->score(denominator,gen,len);

		//delete (*sit).second;
		//to_return << (*sit)->print_full() << endl;
		//delete (*sit);
		delete (*sit).second;
	}
	if(BISULFITE)
		max.score_max_bs(denominator,gen,len,search,to_return);
	if(ATOG)
		max.score_max_atog(denominator,gen,len,search,to_return);

	pthread_mutex_unlock(&lock);
	
	to_return << unique.size() << "\t";
	to_return << max.print_full_w_score(denominator,gen);
	
	//string str = to_return.str();
	//str = "\nHighest Score:\n";
	//string str = "(" << unique.size() << ")" << "\t";
	//str += max.print_full_w_score(denominator);

	if(unique.size() == 0)
		bad_files++;
	
	return to_return.str();
	//return str;
}

string print(Genome &gen, vector<unsigned long> &pos, long begin, int size) {
	string to_return = "";
	
	vector<unsigned long>::iterator sit;
	
	//cout << "matches: " << pos.size() << endl;
	for(sit = pos.begin(); sit != pos.end(); ++sit) {
		to_return += gen.GetString(*sit - begin, size) + "\t";
	}
	
	return to_return;
}


/* GetPWM is used for getting a PWM matrix representing the match (m) and mismatch (mm)
 * penalties.
 * The input is a file, which should be a 4x4 matrix of the format:
 * 				 A   C   G   T
 * 			A    m  mm  mm  mm
 *			C   mm   m  mm  mm
 * 			G   mm  mm   m  mm
 * 			T   mm  mm  mm   m
 * For ease in reading, the letters defining the rows and columns are disregarded.
 *
 *
 * Note:  GetPWM will modify the global ALIGN_SCORES variable.
 */
void getPWM(const char* fn) {
	float temp_ALIGN_SCORES[5][4];

	int THIS_BUFFER = 100;
	
	ifstream in;
	in.open(fn);
	char temp_chr[THIS_BUFFER];
	bool contains_labels=false;
	float a,c,g,t;
	char label[THIS_BUFFER];
	char alabel[THIS_BUFFER];
	char clabel[THIS_BUFFER];
	int count=0;
	int num_lines=4;

	// Get all the lines of input
	while(in.getline(temp_chr,THIS_BUFFER) && count < num_lines) {
		if(sscanf(temp_chr," %s %s %s %s",alabel,clabel,label,label) == 4 
				&& (tolower(*alabel) == 'a')
				&& (tolower(*clabel) == 'c') ) {
			if(VERBOSE > 1)
				printf("Matched here: %s\n",temp_chr);
			in.getline(temp_chr,THIS_BUFFER);
			contains_labels=true;
		}
		if(VERBOSE > 1)
			cout << "Line: " << temp_chr << endl;

		if(contains_labels) {
			if(sscanf(temp_chr,"%s %f %f %f %f",label,&a,&c,&g,&t) != 5) {
				char* error = new char[THIS_BUFFER];
				strcat(error,"Error in Score File: ");
				strcat(error,temp_chr);
				throw(error);
			}
		}
		else {
			if(sscanf(temp_chr,"%f %f %f %f",&a,&c,&g,&t) != 4) {
				char* error = new char[THIS_BUFFER];
				strcat(error,"Error in Score File: ");
				strcat(error,temp_chr);
				throw(error);
			}
		}
		
		temp_ALIGN_SCORES[count][0] = (float)a;
		temp_ALIGN_SCORES[count][1] = (float)c;
		temp_ALIGN_SCORES[count][2] = (float)g;
		temp_ALIGN_SCORES[count][3] = (float)t;

		count++;
	}
	if(count < num_lines) {
		throw("Error in Score File:  Not enough lines");
	}

	cout << "Count: " << count << "\tNum lines: " << num_lines << endl;

#ifdef DEBUG
	//for testing purposes...
	if(VERBOSE > 1) {
		cout << "Matrix: " << endl;
		cout << "\t0\t1\t2\t3" << endl;
		for(int i=0; i<5; i++) {
			cout << i << "\t";
			for(int j=0; j<4; j++) 
				cout << temp_ALIGN_SCORES[i][j] << "\t";
			cout << endl;
		}
		cout << endl;
		throw("Passed.");
	}
#endif

	for(unsigned int i=0; i<4; i++) {
		ALIGN_SCORES[(unsigned int)'a'][i] = temp_ALIGN_SCORES[0][i];
		ALIGN_SCORES[(unsigned int)'c'][i] = temp_ALIGN_SCORES[1][i];
		ALIGN_SCORES[(unsigned int)'g'][i] = temp_ALIGN_SCORES[2][i];
		ALIGN_SCORES[(unsigned int)'t'][i] = temp_ALIGN_SCORES[3][i];
	}
}

void printPWM(Read* read) {
	printf("Length: %d\n",read->length);
	for(unsigned int i=0; i<read->length; i++) {
		printf("%f %f %f %f\n",read->pwm[i][0],read->pwm[i][1],read->pwm[i][2],read->pwm[i][3]);
	}
}

int main(const int argc, const char* argv[]) {
	//Print command line args
	cerr << "This is GNUMAP, Version 1.0, for public and private use." << endl;

	memset(gen_CONVERSION,4,256);
	gen_CONVERSION[(int)'a'] = gen_CONVERSION[(int)'A'] = 0;
	gen_CONVERSION[(int)'c'] = gen_CONVERSION[(int)'C'] = 1;
	gen_CONVERSION[(int)'g'] = gen_CONVERSION[(int)'G'] = 2;
	gen_CONVERSION[(int)'t'] = gen_CONVERSION[(int)'T'] = 3;
	gen_CONVERSION[(int)'n'] = gen_CONVERSION[(int)'N'] = 4;
	gen_CONVERSION[(int)'\n'] = gen_CONVERSION[(int)'\r'] = 5;
	gen_CONVERSION[(int)'\0'] = 6;
	gen_CONVERSION[(int)'>'] = 7;

	memset(bs_CONVERSION,6,256);
	bs_CONVERSION[(int)'a'] = bs_CONVERSION[(int)'A'] = 0;
	bs_CONVERSION[(int)'c'] = bs_CONVERSION[(int)'C'] = 1;
	bs_CONVERSION[(int)'g'] = bs_CONVERSION[(int)'G'] = 2;
	bs_CONVERSION[(int)'t'] = bs_CONVERSION[(int)'T'] = 3;
	bs_CONVERSION[(int)'n'] = bs_CONVERSION[(int)'N'] = 6;
	bs_CONVERSION[(int)'\n'] = bs_CONVERSION[(int)'\r'] = 4;
	bs_CONVERSION[(int)'\0'] = 5;
	bs_CONVERSION[(int)'>'] = 7;

	memset(INT2BASE,'?',16);
	INT2BASE[0] = 'a';
	INT2BASE[1] = 'c';
	INT2BASE[2] = 'g';
	INT2BASE[3] = 't';
	INT2BASE[4] = 'n';


#ifdef DEBUG
	printf("int\tbs\tgen\n");
	for(int i=0; i<256; i++) {
		printf("%d\t%c\t%d\t%d\n",i,(char)i,bs_CONVERSION[i],gen_CONVERSION[i]);
	}
	printf("int2base\n");
	for(int i=0; i<16; i++) {
		printf("%d\t%c\n",i,INT2BASE[i]);
	}
#endif

	cout << endl << "Command Line Arguments:  ";
	for(int i=0; i<argc; i++)
		cout << argv[i] << ' ';
	cout << endl;
	
	VERBOSE = 0;

	if(argc == 1) //means only the bin/gnumap parameter
		usage(0,"");

	if(argc < 2) {
		usage(1, "Need at least a genome and a file to map.");
		exit(1);
	}

	int rc = ParseCmdLine(argc,argv);
	// an error occurred during option processing
	if(rc != 0) {
		GetParseError(cerr, argv);
		return -1;
	}


	//set up alignment matrix

	for(unsigned int i=0; i<256; i++)
		for(unsigned int j=0; j<4; j++)
			ALIGN_SCORES[i][j] = GAP;
	ALIGN_SCORES[(unsigned int)'a'][0] = MATCH;
	ALIGN_SCORES[(unsigned int)'a'][1] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'a'][2] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'a'][3] = MISMATCH;

	ALIGN_SCORES[(unsigned int)'c'][1] = MATCH;
	ALIGN_SCORES[(unsigned int)'c'][0] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'c'][2] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'c'][3] = MISMATCH;

	ALIGN_SCORES[(unsigned int)'g'][2] = MATCH;
	ALIGN_SCORES[(unsigned int)'g'][0] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'g'][1] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'g'][3] = MISMATCH;

	ALIGN_SCORES[(unsigned int)'t'][3] = MATCH;
	ALIGN_SCORES[(unsigned int)'t'][0] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'t'][1] = MISMATCH;
	ALIGN_SCORES[(unsigned int)'t'][2] = MISMATCH;

/*
	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			if(i == j)
				ALIGN_SCORES[i][j] = MATCH;
			else
				ALIGN_SCORES[i][j] = MISMATCH;
		}
		ALIGN_SCORES[4][i] = GAP;
	}
*/
	//determine the size of the hash
	HASH_SIZE = (unsigned int) pow(4.0,MER_SIZE);

	/* Manage errors that might occur */	
	// If they don't include a fasta file for sequences
	// If they don't tell us how long the sequences are
	if( (seq_file == NULL) )
		usage(1, "Specify a single file e.g., sequences.fa\n");
	if(!genome_file)
		usage(1, "Specify a genome to map to with the -g flag\n");

	if(pos_matrix != NULL) {
		try {
			getPWM(pos_matrix);
		}
		catch(const char* err) {
			cerr << "ERROR: \n\t" << err << endl;
			return -1;
		}
	}
	
	ostringstream params;
	params << "Parameters: " << endl;
	params << "\tVerbose: " << VERBOSE << endl;
	params << "\tGenome file(s): " << genome_file << endl;
	params << "\tOutput file: " << output_file << endl;
	params << "\tSequence file: " << seq_file << endl;
	if(!perc)
		params << "\tAlign score: " << ALIGN_SCORE << endl;
	else
		params << "\tAlign percentage: " << ALIGN_SCORE*100 << "%" << endl;
	params << "\tNumber of threads: " << NUM_THREADS << endl;
	if(MAX_HASH_SIZE > 0)
		params << "\tLargest Hash Size: " << MAX_HASH_SIZE << endl;
	params << "\tMer size: " << MER_SIZE << endl;
	if(pos_matrix != NULL)
		params << "\tUsing User-Defined Alignment Scores" << endl;
	else 
		params << "\tUsing Default Alignment Scores" << endl;
	params << "\tGap score: " << GAP << endl;
	params << "\tMaximum Gaps: " << MAX_GAP << endl;
	if(adaptor) {
		params << "\tUsing Adaptor Sequence: " << adaptor << endl;
	}
	if(BISULFITE) {
		PRINT_FULL = true;
		params << "\tUsing BISULFITE conversion" << endl;
		bs_CONVERSION[(int)'c'] = bs_CONVERSION[(int)'C'] = 3;	//change it so every 'c' will be *hashed* as a 't'
		//ALIGN_SCORES[3][1] = MATCH;
		ALIGN_SCORES[1][3] = MATCH;
	}
	if(ATOG) {
		PRINT_FULL = true;
		params << "\tUsing A to G conversion" << endl;
		bs_CONVERSION[(int)'a'] = bs_CONVERSION[(int)'A'] = 2;	//change it so every 'c' will be *hashed* as a 't'
		//ALIGN_SCORES[2][0] = MATCH;
		ALIGN_SCORES[0][2] = MATCH;
	}

#ifdef DEBUG
	printf("Alignment scores:\n");
	for(int i=0; i<5; i++) {
		for(int j=0; j<4; j++) {
			printf("%f\t",ALIGN_SCORES[i][j]);
		}
		printf("\n");
	}
#endif

	if(VERBOSE) {
		cerr << params.str() << endl;
	}
	
	pthread_t pthread_hand[NUM_THREADS];
	void *pthread_ret[NUM_THREADS];
	pthread_mutex_init(&lock,NULL);
	
	if(VERBOSE)
		cerr << "Hashing the genome." << endl;
	Genome gen; 
	try {
		gen.use(genome_file);
		gen.hash_and_store();
	}
	catch(const char* err) {
		cerr << "ERROR: \n\t" << err << endl;
		return -1;
	}
	catch(bad_alloc&) {
		cerr << "ERROR: \n\tUnable to allocate enough memory.\n\t(Pick a smaller mer size?)" << endl;
		return -1;
	}
	catch(exception &e) {
		cerr << "ERROR: \n\t" << e.what() << endl;
		return -1;
	}
	catch(...) {
		cerr << "ERROR: \n\tUnknown error." << endl;
		return -1;
	}

	if(VERBOSE > 1) 
		cerr << "\nTime to hash: " << (1.0*clock())/CLOCKS_PER_SEC << endl;			

	SeqReader sr(seq_file);
	unsigned int total_seqs = 0;
	ofstream of(output_file);
	of << params.str() << endl;
	of << "Consensus\tAlignScoreWithSelf\tMaxAlignScore\tNumUniqueMatches\t"
	   << "MAX:AlignScore\tMAX:LogAlignScore\tMAX:ScoreAddedToPosition\tMAX:Sequence\t"
	   << "MAX:Locations...\n";

	
	//Here is where we'll begin the process of matching and writing to the 
	//output file.
	vector<string> file_vec = sr.GetSeqFileVector();
	
	vector<string>::iterator file_it = file_vec.begin();
	if(VERBOSE) 
		cerr << "Matching " << sr.GetNumSeqFiles() << " file(s):" << endl;
		
	//we'll iterate through all of the files that were included on the commandline.
	for(; file_it!=file_vec.end(); ++file_it) {
		
		if(VERBOSE)
			cerr << "Matching sequences of: " << *file_it << endl;
		try {
			//SequencReader will throw an error if the files are not in the proper
			//format.  This is to catch the error.
			sr.use(*file_it);		
		}
		catch(const char* err) {
			cerr << "ERROR: \n\t" << err << endl;
			if(VERBOSE) 
				cerr << "\tIn file: " << *file_it << endl;
			continue;
		}
		catch(...) {
			cerr << "ERROR: \n\tImproper sequence length or file." << endl;
			cerr << "\tIn file " << *file_it << endl;
			continue;
		}		
		
		unsigned int num_seqs = sr.GetNumSeqs();
		total_seqs+=num_seqs;
		if(VERBOSE)
			cerr << "Total Sequences: " << num_seqs << endl;

		unsigned int each_thread = num_seqs/NUM_THREADS + 1;

#ifdef DEBUG
		if(VERBOSE > 1)
			cout << "Sequences for each thread: " << each_thread << endl;
#endif
		
		//Here is where we set the threading information.
		pthread_setconcurrency(NUM_THREADS);
		pthread_mutex_init(&lock, NULL);
		pthread_attr_init(&attr);
		pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

		//An array of pointers so we can delete them when we're done.
		Read** seq_ptr[NUM_THREADS];
		thread_opts* t_o_ptr[NUM_THREADS];
		 
		//Give each thread an equal amount of sequences.
		for(int i=0; i<NUM_THREADS; i++) {
						
			//We'll create a new vector pointer here.  Will delete it later.
			//vector<vector<double> > *seqs_to_match = new vector<vector<double> >[each_thread];
			Read** seqs_to_match = new Read*[each_thread];
			//Read** seqs_to_match = (Read**)malloc(sizeof(Read*) * each_thread);
			seq_ptr[i] = seqs_to_match;
			
			for(unsigned int j=0; j<each_thread; j++) {
				seqs_to_match[j] = sr.GetNextSequence();
			}
					

			/*
			printf("**************Printing PWMS*********\n");
			for(unsigned int j=0; j<each_thread; j++) {
				if(!seqs_to_match[j])
					continue;
				printf("\n%d: ",j);
				printPWM(seqs_to_match[j]);
			}
			//*/

			//The parameters to pass to the thread
			thread_opts* t_o = new thread_opts(gen,seqs_to_match,each_thread);
			t_o_ptr[i] = t_o;

			//pthread_create(&pthread_hand[i], &attr, thread_print, (void*)&t_o);
			pthread_create(&pthread_hand[i], NULL, thread_print, (void*)t_o);
		}

		//wait for all the threads to join back together...
		for (int i=0; i<NUM_THREADS; i++) {
			pthread_join(pthread_hand[i], &pthread_ret[i]);
		}
		
		//delete the threads and the data so we can reserve memory.
		for(int i=0; i<NUM_THREADS; i++) {
			delete[] seq_ptr[i];
			delete t_o_ptr[i];
		}
		
		//print their return values to the file.
		for(int i=0; i<NUM_THREADS; i++) {
			of << ((ostringstream*)pthread_ret[i])->str();
			delete (ostringstream*)pthread_ret[i];
		}
	
		if(VERBOSE) 
			cerr << endl;
	}
	
	
	if(VERBOSE)
		cerr << "\nPrinting output." << endl;
		
	//Print to the .sgr file
	gen.PrintFinal(output_file);
	
	ostringstream stats;
	stats 	<< "\nFinished." << endl 
			<< "\tTotal Time: " << (1.0*clock())/CLOCKS_PER_SEC << " seconds." << endl
			<< "\tMatched " << total_seqs << " sequences." << endl
			<< "\tSequences not matched: " << bad_files << endl
			<< "\tOutput written to " << output_file << endl;
						
	cerr << stats.str();
	of << stats.str();
	of.close();
		
	return 0;
}

void* thread_print(void* t_opts) {
	//need to create a pointer so we can pass it back...
	ostringstream* of = new ostringstream;
	//cout << "New Thread" << endl;
	unsigned int num_seqs = ((thread_opts*)t_opts)->num_seqs;
	unsigned int progress = num_seqs/100;
	if(progress < 1)
		progress = 1;
		
	for(unsigned int k=0; k < num_seqs; k++) {
		//vector<vector<double> > temp_vect = ((thread_opts*)t_opts)->seqs[k];
		Read* temp_read = ((thread_opts*)t_opts)->seqs[k];
		if(!temp_read) {
			//cout << "Found end" << endl;
			//break;
			continue;
		}
		
		string consensus = GetConsensus(*temp_read);

		if(VERBOSE)
			if( (k % progress ) == 0) {
				cerr << "\r" << (int)(((double)k/num_seqs)*100) << "% complete"; cerr.flush();
			}

		*of << print_top_matches( ((thread_opts*)t_opts)->gen, *temp_read, consensus) << endl;

		//done with the read.  Delete it.
		for(unsigned int i=0; i<temp_read->length; i++) {
			delete[] temp_read->pwm[i];
		}
		delete[] temp_read->pwm;
		delete temp_read;
	}		 
	
	return (void*)of;
}


int ParseCmdLine(const int argc, const char* argv[]) {
	
	bool found_other = false;

	for(int i=1; i<argc; i++) {
		if(*argv[i] == '-') { //check for normal
			int set_ret;

			if(*((argv[i])+1) == '-')  { //check for extended
				//printf("Found extended\n");
				set_ret = set_arg_ext(argv[i],argv[i+1],i);
			}
			else {
				set_ret = set_arg(argv[i],argv[i+1],i);
				i++;	//increment past the next expression
			}
			
			if(set_ret != 0) {
				this_cmd.errnum = set_ret;
				this_cmd.errpos = i;
				return -1;
			}

		}
		else {
			//if we've already found the the other cmd arg, it doesn't work. 
			if(found_other) {
				this_cmd.errnum = INVALID_ARG;
				this_cmd.errpos = i;
				return INVALID_ARG;
			}
			else {	//set the other arg to this pointer
				seq_file = argv[i];
			}
		}
	}

	return 0;

}

void GetParseError(ostream &os, const char* argv[]) {
	if(this_cmd.errnum == PARSE_ERROR) {
		os << "Irregular Parameter in: " << argv[this_cmd.errpos] << endl;
	}
	else if(this_cmd.errnum == NO_MATCHING_ARG) {
		os << "No matching arg in: " << argv[this_cmd.errpos] << endl;
	}
	else if(this_cmd.errnum == INVALID_ARG) {
		os << "Please specify only a single sequence file, or multiple files in \"quotations\" ("
		   << argv[this_cmd.errpos] << ")" << endl;
	}
	else {
		os << "Unknown error " << this_cmd.errnum << " in: " << argv[this_cmd.errpos] << endl;
	}

}


int set_arg(const char* param, const char* assign, int &count) {
	double temp_dbl;

	switch(*(param+1)) {	//switch on the first character
		case 'g':
			genome_file = assign;
			break;
		case 'l':
			if(sscanf(assign,"%d",&SEQ_LENGTH) < 1)
				return PARSE_ERROR;
			break;
		case 'o':
			output_file = assign;
			break;
		case 'a':
			//if(sscanf(assign,"%lf",&ALIGN_SCORE) < 1)
			if(sscanf(assign,"%lf",&temp_dbl)< 1)
				return PARSE_ERROR;
			//	return (PARSE_ERROR * (int)(*param));
			ALIGN_SCORE = (float)temp_dbl;
			break;
		case 'p':
			count--;
			perc = true;
			break;
		case 'v':
			if(sscanf(assign,"%d",&VERBOSE) < 1)
				return PARSE_ERROR;
			break;
		case 'c':
			if(sscanf(assign,"%d",&NUM_THREADS) < 1)
				return PARSE_ERROR;
			break;
		case 'm': 
			if(sscanf(assign,"%d",&MER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'B':
			if(sscanf(assign,"%d",&BUFFER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'T':
			if(sscanf(assign,"%d",&MAX_MATCHES) < 1)
				return PARSE_ERROR;
			break;
		case 'h':
			if(sscanf(assign,"%d",&MAX_HASH_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'S':
			pos_matrix = assign;
			break;
		case 'G':
			//if(sscanf(assign,"%lf",&GAP) < 1)
			if(sscanf(assign,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			GAP = (float)temp_dbl;
			break;
		case 'M':
			if(sscanf(assign,"%d",&MAX_GAP) < 1)
				return PARSE_ERROR;
			break;
		case 'A':
			adaptor = new char[strlen(assign)];
			for(unsigned int i=0; i<strlen(assign); i++)
				adaptor[i] = tolower((char)assign[i]);
			//adaptor = assign;
			break;
		case '0':
			count--;
			PRINT_FULL = 1;
			break;
		case 'b':
			count--;
			BISULFITE = 1;
			break;
		case 'd':
			count--;
			ATOG = 1;
			break;
		case '?':
			usage(0,"");
		default:
			return PARSE_ERROR;
	}

	return 0;

}


int set_arg_ext(const char* param, const char* assign, int &count) {
	double temp_dbl;
	char which=' ';
	int adjust=0;

	if(strncmp(param+2,"genome",6) == 0) {	//check to see if it's the genome
		which = 'g';
		adjust = 8;	//the size of "genome" plus 2
	}
	else if(strncmp(param+2,"length",6) == 0) {
		which = 'l';
		adjust = 8;
	}
	else if(strncmp(param+2,"output",6) == 0) {
		which = 'o';
		adjust = 8;
	}
	else if(strncmp(param+2,"align_score",11) == 0) {
		which = 'a';
		adjust = 13;
	}
	else if(strncmp(param+2,"percent",7) == 0) {
		which = 'p';
		adjust = 9;
	}
	else if(strncmp(param+2,"verbose",7) == 0) {
		which = 'v';
		adjust = 9;
	}
	else if(strncmp(param+2,"num_proc",8) == 0) {
		which = 'c';
		adjust = 10;
	}
	else if(strncmp(param+2,"mer_size",8) == 0) {
		which = 'm';
		adjust = 10;
	}
	else if(strncmp(param+2,"buffer",6) == 0) {
		which = 'B';
		adjust = 8;
	}
	else if(strncmp(param+2,"max_match",8) == 0) {
		which = 'T';
		adjust = 10;
	}
	else if(strncmp(param+2,"max_hash",9) == 0) {
		which = 'h';
		adjust = 11;
	}
	else if(strncmp(param+2,"subst_file",8) == 0) {
		which = 'S';
		adjust = 10;
	}
	else if(strncmp(param+2,"gap_penalty",11) == 0) {
		which = 'G';
		adjust = 13;
	}
	else if(strncmp(param+2,"max_gap",7) == 0) {
		which = 'M';
		adjust = 9;
	}
	else if(strncmp(param+2,"adaptor",7) == 0) {
		which = 'A';
		adjust = 9;
	}
	else if(strncmp(param+2,"print_full",10) == 0) {
		which = '0';
		adjust = 12;
	}
	else if(strncmp(param+2,"bs_seq",6) == 0) {
		which = 'b';
		adjust = 8;
	}
	else if(strncmp(param+2,"a_to_g",6) == 0) {
		which = 'd';
		adjust = 8;
	}
	else if(strcmp(param+2,"help") == 0) {
		usage(0,""); //usage will exit immediately
	}
	else 
		return NO_MATCHING_ARG;

	param += ++adjust;

	switch(which) {	//switch on the first character
		case 'g':
			genome_file = param;
			break;
		case 'l':
			if(sscanf(param,"%d",&SEQ_LENGTH) < 1)
				return PARSE_ERROR;
			break;
		case 'o':
			output_file = param;
			break;
		case 'a':
			//if(sscanf(param,"%lf",&ALIGN_SCORE) < 1)
			if(sscanf(param,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			//	return (PARSE_ERROR * (int)(*param));
			ALIGN_SCORE = (float)temp_dbl;
			break;
		case 'p':
			count--;
			perc = true;
			break;
		case 'v':
			if(sscanf(param,"%d",&VERBOSE) < 1)
				return PARSE_ERROR;
			break;
		case 'c':
			if(sscanf(param,"%d",&NUM_THREADS) < 1)
				return PARSE_ERROR;
			break;
		case 'm': 
			if(sscanf(param,"%d",&MER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'B':
			if(sscanf(param,"%d",&BUFFER_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'T':
			if(sscanf(param,"%d",&MAX_MATCHES) < 1)
				return PARSE_ERROR;
			break;
		case 'h':
			if(sscanf(param,"%d",&MAX_HASH_SIZE) < 1)
				return PARSE_ERROR;
			break;
		case 'S':
			pos_matrix = param;
			break;
		case 'G':
			//if(sscanf(param,"%lf",&GAP) < 1)
			if(sscanf(param,"%lf",&temp_dbl) < 1)
				return PARSE_ERROR;
			GAP = (float)temp_dbl;
			break;
		case 'M':
			if(sscanf(param,"%d",&MAX_GAP) < 1)
				return PARSE_ERROR;
			break;
		case 'A':
			adaptor = new char[strlen(assign)];
			for(unsigned int i=0; i<strlen(assign); i++)
				adaptor[i] = tolower((char)param[i]);
			//adaptor = (char*)param;
			break;
		case '0':
			count--;
			PRINT_FULL = 1;
			break;
		case 'b':
			count--;
			BISULFITE = 1;
			break;
		case 'd':
			count--;
			ATOG = 1;
			break;
		default:
			return NO_MATCHING_ARG;
	}

	return 0;

}
