#include "Genome.h"

/*
 * Constructor.  Creates new Genome with filename.
 * fn - the filename of the file to be parsed into the genome.
 */
Genome::Genome(const char* fn) :
	offset(0), bin_offset(0) {
	gen_piece = new unsigned char[BUFFER_SIZE+1];
	Init(fn);
	reader = new Reader();
}

Genome::Genome() :
	offset(0), bin_offset(0), reader(NULL) {
	gen_piece = new unsigned char[BUFFER_SIZE+1];

}

Genome::~Genome() {
	delete reader;
	delete[] gen_piece;
} 

void Genome::Init(const char* fn) {
	char files[strlen(fn)];
	strcpy(files,fn);
	
	const char* delims = " \"\n";
	char* token = strtok(files,delims);
	if(VERBOSE > 1)
		cout << "Found:" << endl;

	while(token != NULL) {
		if(VERBOSE)
			cout << "\t" << token << endl;
		string temp_str = token;
		names.push_back(pair<string,unsigned long>(temp_str,0));
		
		token = strtok(NULL,delims);
	}

	//Push a dummy string on the end so we can know when to end...
	names.push_back(pair<string,unsigned long>("end",0));

	//clear the *_spots files
	for(int i=0; i<4; i++) {
		reads[i] = 0;
		gens[i] = 0;
	}

}

void Genome::use(const char* fn) {
	if(reader != NULL) {
		cout << "non-null reader" << endl;
		delete reader;
	}
	Init(fn);
	reader = new Reader();
}

void Genome::print_to_file(hash_map<unsigned int, vector<unsigned long> > & gh, char* ofn) { 
	printf("Printing to file: %s\n",ofn);
	int SEQ_LENGTH = 35; //This is an archaic #def'd element that needed to be removed.
	ofstream out;
	out.open(ofn,ios_base::out);
	
	bin_seq bs;
	
	unsigned int i;
	
	for(i=0; i<HASH_SIZE; ++i) {
		if(gh[i].size() == 0)
			continue;
		out << i << "\t" << gh[i].size();
		vector<unsigned long>::iterator sit;
		unsigned int j;
		out << '\t' << bs.hash2str(i,MER_SIZE);
		for(sit = gh[i].begin(), j=0; j<gh[i].size(); j++, ++sit)
			out << '\t' << *sit << "[" << GetPos(*sit,true) << "] :" << GetString(*sit,SEQ_LENGTH);
		out << endl;
	}
	
	out.close();
}

void Genome::hash_and_store() {

	gen_hash.set_size(HASH_SIZE);

	
	// We dont need to use a set--the members are already sorted.  And a vector would have
	// faster random access than a list, which is a plus.
	// hash_map<unsigned int, vector<unsigned int> > gen_hash;
	
	bin_seq bs;	

	unsigned long counter = 0;
	unsigned long int chars_stored = 0;
	unsigned long int j_pos = chars_stored-0;
	unsigned int j;

	for(j=0; j<names.size()-1; j++) {
		reader->use(names[j].first);
		string chr_name = reader->GetName();
		//unsigned long j_pos = counter;
		j_pos = chars_stored;
					
		unsigned long file_size = reader->GetFileSize();
		//cout << "File Size: " << file_size << endl;
		bool keep_reading = true;
		unsigned int i;

		while(keep_reading) { 
	
			keep_reading = reader->read(gen_piece);
	
			//If we're not at the end of the file, then we can throw it the whole chunk,
			//without the extra MER_SIZE bases at the end.
			//When we come to the end of the file, we can give it everything.
			//We'll also return the number of characters stored.
			if(keep_reading) {
				chars_stored += store(gen_piece,BUFFER_SIZE-MER_SIZE);
				//printf("keeping reading\n");
			}
			else
				chars_stored += store(gen_piece,BUFFER_SIZE);
	 
			for(i=0; i < BUFFER_SIZE-MER_SIZE; i++) {
					
				pair<int ,unsigned int> hash = bs.get_hash(i,gen_piece);
	
				if(hash.first >= 0) {	//if it's zero, that's a result of an N character.
					if(hash.first)		//don't want to store N's.
						if(counter % 3 == 0) //This allows us a smaller hash size in total.
							gen_hash[hash.second].push_back(counter);
					counter++;
				}
						
				//if(counter > 15074679)
				//	cout << "Made it..." << endl;

				if( VERBOSE && hash.first && ((counter % (file_size/50) ) == 0)) {
					cout << ".";
					cout.flush();
				}
				
				
				//if hash.first is -2, it means it's a new genome.
				if(hash.first == -2) {
					
					//Find the name of this genome.
					//at position i, we should have the '>'
					char gen_name[100];	//the name can only be 100 chars long
					int gen_pos=0; 
					
					while(gen_piece[++i] != '\n' && i < BUFFER_SIZE) { 
						gen_name[gen_pos] = gen_piece[i];
						gen_pos++;
					}
					gen_name[gen_pos] = '\0'; 
					string str_gen_name = gen_name;
					//printf("Found new genome at pos %u: %s\n",i,gen_name);
					if(VERBOSE > 1) 
						printf("Found new genome: %s\n",gen_name);

					vector<pair<string,unsigned long> >::iterator vit = names.begin();
					//Find the position of this genome name
					for(unsigned int name_pos=0; name_pos < j; name_pos++)
						vit++;
					vit++;	//the vector inserts to the one "previous", but we want to insert
							//after the current name.

					//printf("Inserting: %s, %d at %d before %s, %d ",str_gen_name.c_str(),0,j,(*vit).first.c_str(),(*vit).second);
					//printf("and after %s, %d",(*(vit-1)).first.c_str(),(*(vit-1)).second);
					names.insert(vit,pair<string,unsigned long>(str_gen_name,0));

					//take care of the former name.
					//1. set the name in the genome to have the correct position
					//2. set the j_pos var to be the number of chars we've stored so far.
					names[j] = pair<string,unsigned long>(chr_name,j_pos);
					chr_name = str_gen_name;
					
					counter += hash.second;
					counter -= 1;	//increment the counter the number of bases we've skipped

					j_pos = counter;
					j++;
				}
				

				//if hash.first is -1, it means it's found the end-of-line char.
				if(hash.first == -1 || i >= BUFFER_SIZE-MER_SIZE) {
					//if(hash.first == -1)
					//	printf("Found the eof char\n");
					break;
				}
				//else
				//	counter++;
			}
		}
		if(VERBOSE)
			cout << endl;

		names[j] = pair<string,unsigned long>(chr_name,j_pos);
		
		//there are cases where the end of the line throws off our numbering
		//We'll fix it here.
		unsigned long int end = i;
		int count=0;
		for(unsigned int p=end; gen_piece[p] != '\0'; p++) {
			//if(base2int(gen_piece[p]) <= 4) {
			if(gen_CONVERSION[(int)gen_piece[p]] <= 4) {
				counter++;
				count++;
			}
		}
	}
	j_pos = chars_stored;
	names[j] = pair<string,unsigned long>("end",j_pos);
	
	//print_to_file(gen_hash,"human_hash");
	fix_hash();
	
	//if we are doing the bisulfite sequencing, we want to record what nucleotide
	//comes at each position
	if(BISULFITE) {
		reads[0] = new float[genome.size()];
		reads[1] = new float[genome.size()];
		reads[2] = new float[genome.size()];
		reads[3] = new float[genome.size()];
		gens[1] = new float[genome.size()];
		//gens[3] = new float[genome.size()];
	}
	if(ATOG) {
		reads[0] = new float[genome.size()];
		reads[1] = new float[genome.size()];
		reads[2] = new float[genome.size()];
		reads[3] = new float[genome.size()];
		gens[0] = new float[genome.size()];
		//gens[2] = new float[genome.size()];
	}
#ifdef DEBUG
	for(unsigned int i=0; i<names.size(); i++) {
		printf("Name: %s\tBegin: %ld\n",names[i].first.c_str(),names[i].second);
	}
#endif
}

/*
 * store() will store the genome into a smaller array of BIT64_t portions.
 * Each character is represented by 4 bits (because we require the 4 nucleotides
 * and an additional N character), and can be accessed directly.  
 *
 * One of the problems we have now is the last few characters in gen.  They will
 * appear the second time at the beginning of the genome yet again, throwing off the 
 * numbering system. 
 *
 * @par gen - the piece of the genome to store.
 * @par num - the number of characters to store.
 */
unsigned long int Genome::store(unsigned char* gen, unsigned long int num) {
	unsigned long gen_pos=0;
	unsigned long int num_bases=0;
	
	while(gen_pos<num) {
		GEN_TYPE temp = 0;
		
		if(bin_offset != 0) {
			temp = genome[offset].first;
			genome.pop_back();
		}
			
		for(; (bin_offset<GEN_SIZE) && gen_pos<num; gen_pos++) {	//can fit 16 chars into a BIT64_t 
			//int int_to_add = base2int(gen[gen_pos]);
			int int_to_add = gen_CONVERSION[(int)gen[gen_pos]];
			
			//If we come accross the flag for the end of the stream.
			if(int_to_add == 6) { 
				gen_pos = num;
				//cout << "Found end." << endl;
				break;
			}
			
			//If it's a valid char.
			else if(int_to_add < 5) {
				num_bases++;
				temp <<= 4;
				temp |= int_to_add;
				++bin_offset;
			}
			else if(int_to_add == 7) {	//come accross a new name.
				while(gen[gen_pos] != '\n')
					gen_pos++;	//increment the counter until we've found the end of the name
			}
			else ;//do nothing--it's an invalid character and not an end-of-line
		}
		if(bin_offset != 0)	{ //if we've made any changes to this value...
			genome.push_back(pair<GEN_TYPE,float>(temp,0));
		}
		else ;	//no changes.  Coming out of the previous for-loop
		
		
		//means we've filled it fully--so increment the offset.
		if(bin_offset == GEN_SIZE) {
			offset++;
			bin_offset = 0;
		}


	}
	
	//printf("Just read %ld bases\n",num_bases);
	return num_bases;
}

void Genome::fix_hash() {
	//Means there is no user-defined max-hash.
	if(MAX_HASH_SIZE == 0)
		return;

	for(unsigned int i=0; i<gen_hash.size(); i++)
		if(gen_hash[i].size() > MAX_HASH_SIZE) {
			//printf("Clearing hash at pos %d\n",i);
			gen_hash[i].clear();
		}
}


/*
 * This particular store function is used for the DataGen program--not for genhash. 
 * It is meant to store the genome without needing to hash every member--speeds up the proccess
 * a bit.
 */
void Genome::store() {
	if(VERBOSE)
		cerr << "Storing the genome." << endl;

	gen_hash.set_size(HASH_SIZE);

	
	// We dont need to use a set--the members are already sorted.  And a vector would have
	// faster random access than a list, which is a plus.
	// hash_map<unsigned int, vector<unsigned int> > gen_hash;
	
	bin_seq bs;	

	unsigned long counter = 0;
	unsigned long int chars_stored = 0, j_pos = 0;
	unsigned int j;

	for(j=0; j<names.size()-1; j++) {
		reader->use(names[j].first);
		string chr_name = reader->GetName();
		j_pos = chars_stored;
					
		unsigned long file_size = reader->GetFileSize();
		unsigned int i;
		bool keep_reading = true;
		
		while(keep_reading) { 
	
			keep_reading = reader->read(gen_piece);
	
			//If we're not at the end of the file, then we can throw it the whole chunk,
			//without the extra MER_SIZE bases at the end.
			//When we come to the end of the file, we can give it everything.
			if(keep_reading)
				chars_stored += store(gen_piece,BUFFER_SIZE-MER_SIZE);
			else
				chars_stored += store(gen_piece,BUFFER_SIZE);
	
			//for(unsigned int i=0; i < BUFFER_SIZE-MER_SIZE; i+=3) {
			//for(unsigned int i=0; i < BUFFER_SIZE-MER_SIZE; i++) {
			for(i=0; i < BUFFER_SIZE-MER_SIZE; i++) {
									
				pair<bool,unsigned int> hash = bs.get_hash(i,gen_piece);

				if(hash.first > 0) {
					if(counter % 3 == 0) //This allows us a smaller hash size in total.
						gen_hash[hash.second].push_back(counter);
					counter++;
				}
						
				if( VERBOSE && hash.first && ((counter % (file_size/50) ) == 0)) {
					cout << ".";
					cout.flush();
				}
				
				
				//if hash.first is -2, it means it's a new genome.
				if(hash.first == -2) {
					//Find the name of this genome.
					//at position i, we should have the '>'
					char gen_name[100];	//the name can only be 100 chars long
					int gen_pos=0; 
					while(gen_piece[++i] != '\n') { 
						gen_name[gen_pos] = gen_piece[i];
						gen_pos++;
					}
					gen_name[gen_pos] = '\0'; 
					string str_gen_name = gen_name;

					vector<pair<string,unsigned long> >::iterator vit = names.begin();
					//Find the position of this genome name
					for(unsigned int name_pos=0; name_pos < j; name_pos++)
						vit++;
					vit++;	//the vector inserts to the one "previous", but we want to insert
							//after the current name.

					names.insert(vit,pair<string,unsigned long>(str_gen_name,0));

					//take care of the former name.
					//1. set the name in the genome to have the correct position
					//2. set the j_pos var to be the number of chars we've stored so far.
					names[j] = pair<string,unsigned long>(chr_name,j_pos);
					chr_name = str_gen_name;
					
					counter += hash.second-1;	//increment the counter the number of bases we've skipped
					j_pos = counter;
					j++;
				}
				

				//if hash.first is -1, it means it's found the end-of-line char.
				if(hash.first == -1 || i >= BUFFER_SIZE-MER_SIZE) {
					break;
				}

			}
		}
		if(VERBOSE)
			cout << endl;
		names[j] = pair<string,unsigned long>(chr_name,j_pos);

		//there are cases where the end of the line throws off our numbering.
		//We'll fix it here.

		unsigned int end = i;
		int count=0;
		for(unsigned int p=end; gen_piece[p] != '\0'; p++) {
			//if(base2int(gen_piece[p]) <= 4) {
			if(gen_CONVERSION[(int)gen_piece[p]] <= 4) {
				counter++;
				count++;
			}
		}

	}

	j_pos = chars_stored;
	names[j] = pair<string,unsigned long>("end",j_pos);

#ifdef DEBUG
	for(unsigned int i=0; i<names.size(); i++) {
		printf("Name: %s\tBegin: %ld\n",names[i].first.c_str(),names[i].second);
	}
#endif
}

/*
 * Returns a string from begin to begin+size.
 * the first position in the genome is position 0.
 * @par begin - the beginning of the string to return
 * @par size - the size of the string to return.
 */
string Genome::GetString(const unsigned long begin, const unsigned int size) { 
	string to_return = "";

	/***********************Check for Bounds******************************/
	//We don't want to get a string beyond the end of the chromosome...
	int chr_num=0;
	//											it's already been pre-incremented
	while(names[chr_num].second <= begin && ((unsigned int)chr_num <= names.size()) ) chr_num++;
	//while(names[++chr_num].second <= begin && ((unsigned int)chr_num <= names.size()) );
	
	//means it's not valid--will extend beyond the chromosome file
	if( (names[chr_num].second < begin+size) ) 
		//printf("Extending Beyong the Chromosome\n");
		return "";
	/*********************************************************************/

	unsigned long pos_in_gen = begin/GEN_SIZE; 
	unsigned int temp_size = size;
	
	GEN_TYPE t_temp = genome[pos_in_gen].first;

	//cout << "\tbegin: " << begin << endl;
	//cout << "\tpos_in_gen: " << pos_in_gen << endl;
	//cout << "\tt_temp: " << t_temp << endl;
		
	//if(VERBOSE > 3)
	//	cout << "getting position: " << t_temp << endl;
	
	unsigned int count=0;

	while(count < size) {

		string temp = "";
	
		if(count == 0) {	//get the last few characters off the BIT64_t
			unsigned long bit_offset = (begin%GEN_SIZE);

			//for(unsigned long j=0; j<GEN_SIZE-bit_offset; j++) {
			for(unsigned int j=GEN_SIZE-bit_offset; j>0; j--) {
				//shift 0xf left 4 bases for each j.
				//then we don't need to reverse the string.
				//printf("Shifting by: %d = %lld\n",4*(j-1),(t_temp>>(4*(j-1))) & 0xf);
				//temp += int2base(t_temp & 0xf);
				//temp += int2base(0xf & (t_temp >> (4*(j-1))));
				temp += INT2BASE[0xf & (t_temp >> (4*(j-1)))];
				//t_temp >>= 4;
				count++;
			}
			//temp = reverse(temp);
			if(temp.size() > size)
				return temp.substr(0,size);
		}
		else {	//we now want the first offset characters off the BIT64_t.
			//for(unsigned int j=0; j<GEN_SIZE; j++) {
			for(unsigned int j=GEN_SIZE; j>0; j--) {
				
				//printf("Shifting by: %d = %lld\n",4*(j-1),(t_temp>>(4*(j-1))) & 0xf);
				//temp += int2base(t_temp & 0xf);
				//temp += int2base((t_temp >> (4*(j-1))) & 0xf);
				temp += INT2BASE[(t_temp >> (4*(j-1))) & 0xf];
				//t_temp >>= 4;
			}
			//temp = reverse(temp);
			temp = temp.substr(0,temp_size);
			count += temp.size();
		}
		to_return += temp;
		
		temp_size = size - count;
		
		t_temp = genome[++pos_in_gen].first;
	}

	return to_return;
}

inline int Genome::base2int(char c) {
	switch(c) {
		case 'A' : return 0;
		case 'C' : return 1;
		case 'G' : return 2;
		case 'T' : return 3;
		case 'a' : return 0;
		case 'c' : return 1;
		case 'g' : return 2;
		case 't' : return 3; 
		case 'N' : return 4;
		case 'n' : return 4;
		case '\n' : return 5;
		case '\r' : return 5;
		case '\0' : return 6;	//6 is the flag for the end of the genome stream.
		case '>' : return 7;
		default: return 4;	//we want invalid characters to be represented by "N"
	}
	return 5;	//5 is an invalid character.
} 

inline char Genome::int2base(int i) {
	//printf("i: %d\n",i);
	switch(i) {
		case 0: return 'a';
		case 1: return 'c';
		case 2: return 'g';
		case 3: return 't';
		case 4: return 'n';
	}
	return '?';
}

inline string Genome::reverse(string &s) {
	string temp = "";
	int len = s.length();
	
	for(int i=1; i<=len; i++)
		temp += s[len-i];
	
	return temp;
}

vector<unsigned long>& Genome::GetMatches(unsigned int hash) {
	return gen_hash[hash];
}

void Genome::AddScore(const unsigned long &pos, const float &amt) {
	//float pos_temp = pos / BIN_SIZE;
	//pos_temp *= BIN_SIZE;
	unsigned long pos_temp = pos;
	genome[pos_temp/GEN_SIZE].second += amt;
}

void Genome::AddSeqScore(const unsigned long pos, const float amt, char which) {
	switch(which) {
		case 'a':
			if(reads[0])	//if we've created one, then add to it.  Used for both BISULFITE and ATOG
				reads[0][pos/GEN_SIZE] += amt;
		break;
		case 'c':
			if(reads[1])
				reads[1][pos/GEN_SIZE] += amt;
		break;
		case 'g':
			if(reads[2])
				reads[2][pos/GEN_SIZE] += amt;
		break;
		case 't':
			if(reads[3])
				reads[3][pos/GEN_SIZE] += amt;
		break;
	}
}

void Genome::AddGenScore(const unsigned long pos, const float amt, char which) {
	//genome[pos/GEN_SIZE].second += amt;	//we only want to add it to the genome once.
										//we'll do it while we're adding to the genome sequence files.
							//actually, we're already adding the score of each sequence to the genome.
							//we're doing it a second time here.  bad news, bears.
	switch(which) {
		case 'a':
			if(gens[0])
				//gens[0][pos/GEN_SIZE] += amt;
				gens[0][pos/GEN_SIZE] = 1;
		break;
		case 'c':
			if(gens[1])
				//gens[1][pos/GEN_SIZE] += amt;
				gens[1][pos/GEN_SIZE] = 1;
		break;
		case 'g':
			if(gens[2])
				//gens[2][pos/GEN_SIZE] += amt;
				gens[2][pos/GEN_SIZE] = 1;
		break;
		case 't':
			if(gens[3])
				//gens[3][pos/GEN_SIZE] += amt;
				gens[3][pos/GEN_SIZE] = 1;
		break;
	}
}

string Genome::GetPos(const pair<unsigned long,bool> p_pos) {
	return GetPos(p_pos.first,p_pos.second);
}
string Genome::GetPos(const unsigned long &pos, const bool upstrand) {
	int chr_num = 0;

	while(names[chr_num].second <= pos && ((unsigned int)chr_num < names.size()) )
		chr_num++;
		
	ostringstream to_return;
	
	chr_num--;
	to_return << names[chr_num].first << ":";
	if(upstrand)
		to_return << "+";
	else
		to_return << "-";
	to_return << (pos - names[chr_num].second);
	
	return to_return.str();
}

void Genome::PrintFinal(const char* fn) {
	char* filename = new char[100];
	strcpy(filename,fn);  
	ofstream out(strcat(filename,".sgr"));
	
	for(unsigned int i=0,count=0; i<names.size()-1; i++) {
		for(; count<names[i+1].second; count+=GEN_SIZE) {
			if(genome[count/GEN_SIZE].second > 0)
				out << names[i].first << "\t" << count - (names[i].second) << "\t" << (genome[count/GEN_SIZE].second) << endl;
		}
	}
	out.close();

	for(unsigned i=0; i<4; i++) {
		if(reads[i]) {
			char readfile[100+6];
			strcpy(readfile,fn);

			strcat(readfile,".");
			if(i == 0)
				strcat(readfile,"a");
			else if(i == 1)
				strcat(readfile,"c");
			else if(i == 2)
				strcat(readfile,"g");
			else if(i == 3)
				strcat(readfile,"t");

			strcat(readfile,".sgr");
			out.open(readfile);
			for(unsigned int j=0,count=0; j<names.size()-1; j++) {
				for(; count<names[j+1].second; count+=GEN_SIZE) {
					if(reads[i][count/GEN_SIZE] > 0)
						out << names[j].first << "\t" << count - (names[j].second) << "\t" << reads[i][count/GEN_SIZE] << endl;
					//if(a_spots[count/GEN_SIZE] > 0)
						//out << names[i].first << "\t" << count - (names[i].second) << "\t" << a_spots[count/GEN_SIZE] << endl;
				}

			}
			out.close();
		}

		if(gens[i]) {
			char genfile[100+10];
			strncpy(genfile,fn,100);
			strcat(genfile,".gen.");
			if(i == 0)
				strcat(genfile,"a");
			else if(i == 1)
				strcat(genfile,"c");
			else if(i == 2)
				strcat(genfile,"g");
			else if(i == 3)
				strcat(genfile,"t");

			strcat(genfile,".sgr");
			out.open(genfile);
			for(unsigned int j=0,count=0; j<names.size()-1; j++) {
				for(; count<names[j+1].second; count+=GEN_SIZE) {
					if(gens[i][count/GEN_SIZE] > 0)
						out << names[j].first << "\t" << count - (names[j].second) << "\t" << gens[i][count/GEN_SIZE] << endl;
					//if(a_spots[count/GEN_SIZE] > 0)
						//out << names[i].first << "\t" << count - (names[i].second) << "\t" << a_spots[count/GEN_SIZE] << endl;
				}

			}
			out.close();
		}
	}
	
	delete[] filename;
}

unsigned long Genome::size() {
	return genome.size() * GEN_SIZE;
}

bool Genome::Test(ostream &os) {
	bool success = true;
	 
	unsigned char* genome = new unsigned char[1000000];
	ifstream in("test/short.fa");
	string name;
	getline(in,name);
	in.read((char*)genome,1000000);
	
	//&genome.front() = "actgctgccggtN\ncccgtcggcaacgtgNNNNNgaa\n\0";
	
	try {
		Genome gen0;
		//gen0.use("/data/public/genomes/human/chr2.fa /data/public/genomes/human/chr18.fa /data/public/genomes/human/chr21.fa /data/public/genomes/human/chr22.fa");
		gen0.use("test/short1.fa test/short2.fa test/short3.fa test/short4.fa");
		gen0.hash_and_store();
		printf("Getting string %d at location %s: %s\n",372-122,gen0.GetPos(372-122,false).c_str(),gen0.GetString(372-122,35).c_str());
		printf("Getting string %d at location %s: %s\n",372-122-10,gen0.GetPos(372-122-10,false).c_str(),gen0.GetString(372-122-10,35).c_str());
		printf("Getting string %d at location %s: %s\n",750,gen0.GetPos(750,false).c_str(),gen0.GetString(750,35).c_str());
		printf("Getting string %d at location %s: %s\n",750-10,gen0.GetPos(750-10,false).c_str(),gen0.GetString(750-10,35).c_str());
		printf("Getting string %d at location %s: %s\n",1176,gen0.GetPos(1176,false).c_str(),gen0.GetString(1176,35).c_str());
		printf("Getting string %d at location %s: %s\n",1176-10,gen0.GetPos(1176-10,false).c_str(),gen0.GetString(1176-10,35).c_str());
		//printf("Getting string %d at location %s: %s\n",242985162-34011,gen0.GetPos(242985162-34011,false).c_str(),gen0.GetString(242985162-34011,35).c_str());
		//printf("Getting string %d at location %s: %s\n",242985162-34011-10,gen0.GetPos(242985162-34011-10,false).c_str(),gen0.GetString(242985162-34011-10,35).c_str());


		printf("\n\n***Beginning Next Test\n");
		Genome gen("test/short.fa");

		//gen.store(genome,1000000);
		gen.store();

		string to_reverse = "racecar";
		string to_reverse2 = "a";
		TEST(gen.reverse(to_reverse) == to_reverse);
		TEST(gen.reverse(to_reverse2) == to_reverse2);

		string temp = gen.GetString(8904,15);
		string temp1A = "tattcccgtatccaa";
		os << "temp at 8904: " << temp << endl;
		os << "should be   : " << temp1A << endl;
		TEST(temp.compare(temp1A) == 0);
		temp = gen.GetString(3547,5);
		os << "temp at 3547: " << temp << endl
		   << "should be   : " << "nnnca" << endl;
		TEST(temp == "nnnca");
		
		os << "temp at 57057: " << gen.GetString(57057,9) << endl;
		TEST(gen.GetString(7225,9) == "tctacaaaa");
		TEST(gen.GetString(57057,9) == "acttgatga");
		TEST(gen.GetString(112345,9) == "tgtttactg");

		in.close();
		
		//Genome gen3;
		//gen3.use("test/ecoli.fa");
		//gen3.hash_and_store();
		//os << "string at 151013: " << gen3.GetString(151013,26) << endl;
		os << "**TESTING STORE**\n";
		Genome gen3;
		gen3.use("test/short.fa");
		gen3.store();
		TEST(gen3.GetString(7225,9) == "tctacaaaa");
		TEST(gen3.GetString(57057,9) == "acttgatga");
		TEST(gen3.GetString(112345,9) == "tgtttactg");
		TEST(gen3.GetString(2123345,15) == "caagtgaaacaaagt");
		string sequence = "tactgctgtatcttcagagcctagaataatt";
		TEST(gen3.GetString(13877156,MER_SIZE) == sequence.substr(0,MER_SIZE));
		sequence = "aatgctaac";
		TEST(sequence == gen3.GetString(209879,9))


		//These are for testing the entire program--loading, hashing, and storing.
		os << "**TESTING HASH_AND_STORE**\n";
		Genome gen2;
		gen2.use("test/short.fa");
		gen2.hash_and_store();
		os << "temp at 7225: " << gen2.GetString(7225,9) << endl;
		TEST(gen2.GetString(7225,9) == "tctacaaaa");
		TEST(gen2.GetString(57057,9) == "acttgatga");
		TEST(gen2.GetString(112345,9) == "tgtttactg");
		TEST(gen2.GetString(2123345,15) == "caagtgaaacaaagt");
						  
		sequence = "tactgctgtatcttcagagcctagaataatt";
		TEST(gen2.GetString(13877156,MER_SIZE) == sequence.substr(0,MER_SIZE));
					  
		sequence = "aatgctaac";
		TEST(sequence == gen2.GetString(209879,9))
		
		Genome gen4;
		gen4.use("test/chrI.fa");
		gen4.hash_and_store();
		os << "At position 2047747: " << gen4.GetString(2047747,50) << endl;
		os << "At position 2047650: " << gen4.GetString(2047650,50) << endl;
		
		os << "At position 14155480: " << gen4.GetString(14155490,100) << endl;
		os << "At position 14153950: " << gen4.GetString(14153950,100) << endl;
		
		os << "At position 14155780: " << gen4.GetString(14155780,200) << endl;
		os << "At position 14153680: " << gen4.GetString(14153680,200) << endl;

		os << "At position 13162810: " << gen4.GetString(13162810,50) << endl;
		os << "At position 11142400: " << gen4.GetString(11142400,50) << endl;
		
	
		//MER_SIZE=14;
		//HASH_SIZE = pow(4.0,14);	//HASH_SIZE is set as 4^9

		string test_seq = "accagaaactgtcaagtgatccagctcgag";
		Genome gen5;
		gen5.use("/data/public/genomes/C_elegans/chrI.fa /data/public/genomes/C_elegans/chrII.fa /data/public/genomes/C_elegans/chrIII.fa /data/public/genomes/C_elegans/chrIV.fa");
		//gen5.use("/data/public/genomes/human/chr2.fa /data/public/genomes/human/chr20.fa /data/public/genomes/human/chr21.fa /data/public/genomes/human/chr22.fa");
		gen5.hash_and_store();
		os << "At position NGF (115657200): " << gen5.GetString(115657200,200) << endl;
		os << "At position RXRG (163659640): " << gen5.GetString(163659640,200) << endl;

		Genome gen6;
		gen6.use("/data/public/genomes/C_elegans/chrI.fa /data/public/genomes/C_elegans/chrII.fa /data/public/genomes/C_elegans/chrIII.fa /data/public/genomes/C_elegans/chrIV.fa");
		//gen6.use("/data/public/genomes/human/chr2.fa /data/public/genomes/human/chr20.fa /data/public/genomes/human/chr21.fa /data/public/genomes/human/chr22.fa");
		gen6.store();
		printf("at gen6: %s\nat gen5: %s\n",gen6.GetString(44116093,30).c_str(),
											gen5.GetString(44116093,30).c_str()); 
		TEST(!gen6.GetString(305387122,30).compare(gen5.GetString(305387122,30)));
		printf("%s\n%s\n",gen5.GetPos(44116093,0).c_str(),gen5.GetString(44116093,30).c_str());
		printf("%s\n%s\n",gen6.GetPos(44116093,0).c_str(),gen6.GetString(44116093,30).c_str());

		/*os << "At position 2489000: " << gen4.GetString(2489000,50) << endl;
		os << "At position 2489017: " << gen4.GetString(2489017,50) << endl;
		os << "At position 2489048: " << gen4.GetString(2489048,50) << endl;
		os << "At position 2489049: " << gen4.GetString(2489049,50) << endl << endl;
	//	os << "At position 4457200: " << gen4.GetString(4457200,200) << endl << endl;
		//os << "At position 7744500: " << gen4.GetString(7744500,50) << endl;
		os << "At position 7744416:                           " << gen4.GetString(7744416,30) << endl;
		os << "At position 7744400:           " << gen4.GetString(7744400,50) << endl;
		os << "At position 7744390: " << gen4.GetString(7744390,50) << endl;
		//os << "*At position 7744250: " << gen4.GetString(7744250,50) << endl;
		//os << "*At position 5000000: " << gen4.GetString(5000000,50) << endl << endl;

	//	os << "At position 10485500: " << gen4.GetString(10485500,50) << endl;
	//	os << "At position 10485500: " << gen4.GetString(10485500,300) << endl;
	*/
		
		//ofstream outfile("what_we_got.fa");
		//outfile << gen2.GetString(0,209893+10);
		//outfile.close();
		
		
		
		//ofstream outfile("what_we_got21.fa");
		//outfile << gen3.GetString(1,209893);
		//outfile.close();
	}
	catch(const char* erro) {
		os << "\t********Error! Did you include the proper test files?" << endl;
	}
	catch (...) {
		os << "Unknown Error!" << endl;
	}

	delete[] genome;
	return success;
}
